@@ -1,7 +1,9 @@
 """Client for Kameleoon"""
+
 import asyncio
 import itertools
 import time
+import warnings
 from http.cookies import Morsel
 from typing import cast, Callable, Coroutine, Optional, Tuple, Union, Any, List, Dict
 
@@ -12,7 +14,7 @@ from kameleoon.configuration.data_file import DataFile
 from kameleoon.configuration.feature_flag import FeatureFlag
 from kameleoon.configuration.rule import Rule
 from kameleoon.configuration.rule_type import RuleType
-from kameleoon.configuration.variation import Variation
+from kameleoon import configuration
 from kameleoon.configuration.variation_by_exposition import VariationByExposition
 from kameleoon.helpers.visitor_code import validate_visitor_code
 from kameleoon.hybrid.hybrid_manager_impl import HybridManagerImpl
@@ -21,6 +23,7 @@ from kameleoon.managers.warehouse.warehouse_manager import WarehouseManager
 from kameleoon.data.manager.assigned_variation import AssignedVariation
 from kameleoon.data.manager.visitor import Visitor
 from kameleoon.data.manager.visitor_manager import VisitorManager
+from kameleoon.targeting.targeting_manager import TargetingManager
 from kameleoon.network.access_token_source_factory import AccessTokenSourceFactory
 from kameleoon.network.activity_event import ActivityEvent
 from kameleoon.network.sendable import Sendable
@@ -45,8 +48,6 @@ from kameleoon.helpers.scheduler import Scheduler
 from kameleoon.real_time.real_time_configuration_service import (
     RealTimeConfigurationService,
 )
-from kameleoon.targeting.conditions.sdk_language_condition import SdkLanguageCondition
-from kameleoon.sdk_version import SdkVersion
 
 from kameleoon.exceptions import (
     FeatureVariationNotFound,
@@ -54,13 +55,17 @@ from kameleoon.exceptions import (
     FeatureEnvironmentDisabled,
     SiteCodeIsEmpty,
 )
-from kameleoon.helpers.functions import obtain_hash_double_rule, enum_from_literal
-from kameleoon.targeting.conditions.targeting_condition import TargetingConditionType
+from kameleoon.helpers.functions import obtain_hash_double_rule
 
 __all__ = [
     "KameleoonClient",
 ]
 
+from kameleoon.managers.remote_data.remote_data_manager import RemoteDataManager
+from kameleoon.types.remote_visitor_data_filter import RemoteVisitorDataFilter
+from kameleoon.types.variable import Variable
+from kameleoon.types.variation import Variation
+
 REFERENCE = 0
 X_PAGINATION_PAGE_COUNT = "X-Pagination-Page-Count"
 SEGMENT = "targetingSegment"
@@ -130,9 +135,17 @@ class KameleoonClient:
         self._cookie_manager = CookieManager(config.top_level_domain)
         atsf = AccessTokenSourceFactory(config.client_id, config.client_secret)
         self._network_manager = self._network_manager_factory.create(
-            self.site_code, config.environment, config.default_timeout_second, NetProviderImpl(), atsf, config.logger
+            self.site_code,
+            config.environment,
+            config.default_timeout_second,
+            NetProviderImpl(),
+            atsf,
+            config.logger,
         )
+        self._targeting_manager = TargetingManager(self._visitor_manager, self._data_file)
         self._warehouse_manager = WarehouseManager(self._network_manager, self._visitor_manager, config.logger)
+        self._remote_data_manager = RemoteDataManager(self._network_manager, self._visitor_manager, config.logger)
+        self._visitor_manager.custom_data_info = self._data_file.custom_data_info
         self._thread_event_loop = ThreadEventLoop() if config.multi_threading else None
 
         self._timer: Optional[RepeatTimer] = self._scheduler.timer()
@@ -221,7 +234,10 @@ class KameleoonClient:
         return self._cookie_manager.get_or_add(cookies_readonly, cookies, default_visitor_code)
 
     def set_legal_consent(
-        self, visitor_code: str, consent: bool, cookies: Optional[Dict[str, Morsel[str]]] = None
+        self,
+        visitor_code: str,
+        consent: bool,
+        cookies: Optional[Dict[str, Morsel[str]]] = None,
     ) -> None:
         """
         Sets or updates the legal consent status for a visitor identified by their unique visitor code,
@@ -291,11 +307,12 @@ class KameleoonClient:
                 kameleoon_client.add_data(visitor_code, Interest(1))
         """
         validate_visitor_code(visitor_code)
-        visitor = self._visitor_manager.get_or_create_visitor(visitor_code)
-        visitor.add_data(*args, logger=self._config.logger)
+        self._visitor_manager.add_data(visitor_code, *args, self._config.logger)
         self._config.logger.debug("Successfully added data")
 
-    def track_conversion(self, visitor_code: str, goal_id: int, revenue: float = 0.0) -> None:
+    def track_conversion(
+        self, visitor_code: str, goal_id: int, revenue: float = 0.0, is_unique_identifier: bool = False
+    ) -> None:
         """
         To track conversion, use the track_conversion() method. This method requires visitor_code and goal_id to track
         conversion on this particular goal. In addition, this method also accepts revenue as a third optional argument
@@ -309,13 +326,15 @@ class KameleoonClient:
         :type goal_id: int
         :param revenue: Revenue of the conversion. This field is optional.
         :type revenue: float
+        :param is_unique_identifier: Parameter that specifies whether the visitorCode is a unique identifier.
+        :type is_unique_identifier: bool
         :return: None
         """
         validate_visitor_code(visitor_code)
         self.add_data(visitor_code, Conversion(goal_id, revenue))
-        self.flush(visitor_code)
+        self.flush(visitor_code, is_unique_identifier)
 
-    def flush(self, visitor_code: Optional[str] = None):
+    def flush(self, visitor_code: Optional[str] = None, is_unique_identifier: bool = False):
         """
         Data associated with the current user via add_data() method is not immediately sent to the server.
         It is stored and accumulated until it is sent automatically by the trigger_experiment()
@@ -329,6 +348,8 @@ class KameleoonClient:
 
         :param visitor_code: Unique identifier of the user. This field is mandatory.
         :type visitor_code: Optional[str]
+        :param is_unique_identifier: Parameter that specifies whether the visitorCode is a unique identifier.
+        :type is_unique_identifier: bool
 
         Examples:
 
@@ -349,13 +370,13 @@ class KameleoonClient:
         if visitor_code is not None:
             validate_visitor_code(visitor_code)
             visitor = self._visitor_manager.get_visitor(visitor_code)
-            self.__send_tracking_request(visitor_code, visitor)
+            self.__send_tracking_request(visitor_code, visitor, True, is_unique_identifier)
         else:
             for visitor_code in self._visitor_manager:  # pylint: disable=R1704
                 visitor = self._visitor_manager.get_visitor(visitor_code)
-                self.__send_tracking_request(visitor_code, visitor, False)
+                self.__send_tracking_request(visitor_code, visitor, False, is_unique_identifier)
 
-    def is_feature_active(self, visitor_code: str, feature_key: str) -> bool:
+    def is_feature_active(self, visitor_code: str, feature_key: str, is_unique_identifier: bool = False) -> bool:
         """
         Check if feature is active for a given visitor code
 
@@ -370,6 +391,7 @@ class KameleoonClient:
 
         :param visitor_code: str Unique identifier of the user. This field is mandatory.
         :param feature_key: str Key of the feature flag you want to expose to a user. This field is mandatory.
+        :param is_unique_identifier: bool Parameter that specifies whether the visitorCode is a unique identifier.
         :return: bool Value of the feature that is active for a given visitor_code.
 
 
@@ -400,8 +422,8 @@ class KameleoonClient:
                     # Implement new checkout code here
         """
         try:
-            (_, variation_key) = self.__get_feature_variation_key(visitor_code, feature_key)
-            return variation_key != Variation.Type.OFF.value
+            (_, variation_key) = self.__get_feature_variation_key(visitor_code, feature_key, is_unique_identifier)
+            return variation_key != configuration.Variation.Type.OFF.value
         except FeatureEnvironmentDisabled:
             return False
 
@@ -460,9 +482,7 @@ class KameleoonClient:
         :return: data assosiated with this key, decoded into json
         :rtype: Optional[Any]
         """
-        service: DataService = self._network_manager.get_service(DataService)
-        response = await service.get_remote_data(key, timeout)
-        return response.content
+        return await self._remote_data_manager.get_data(key, timeout)
 
     def get_remote_data(self, key: str, timeout: Optional[float] = None) -> Optional[Any]:
         """
@@ -481,13 +501,16 @@ class KameleoonClient:
         :return: data assosiated with this key, decoded into json
         :rtype: Optional[Any]
         """
-        service: DataService = self._network_manager.get_service(DataService)
-        coro = service.get_remote_data(key, timeout, sync=self.__is_sync_mode())
-        response = cast(Response, self.__make_sync_call_anyway(coro, "get_remote_data"))
-        return response.content
+        coro = self._remote_data_manager.get_data(key, timeout, self.__is_sync_mode())
+        return self.__make_sync_call_anyway(coro, "get_remote_data")
 
     async def get_remote_visitor_data_async(
-        self, visitor_code: str, add_data=True, timeout: Optional[float] = None
+        self,
+        visitor_code: str,
+        add_data=True,
+        timeout: Optional[float] = None,
+        data_filter: Optional[RemoteVisitorDataFilter] = None,
+        is_unique_identifier: bool = False,
     ) -> List[Data]:
         """
         The get_remote_visitor_data_async is an asynchronous method for retrieving custom data for
@@ -502,15 +525,26 @@ class KameleoonClient:
         :param timeout: requests Timeout for request (in seconds). Equals default_timeout in a config file.
         This field is optional.
         :type timeout: Optional[float]
+        :param data_filter: Filter that specifies which data should be retrieved from visits.
+        :type data_filter: RemoteVisitorDataFilter
+        :param is_unique_identifier: Parameter that specifies whether the visitorCode is a unique identifier.
+        :type is_unique_identifier: bool
 
         :return: A list of data assigned to the given visitor.
         :rtype: List[Data]
         """
-        service: DataService = self._network_manager.get_service(DataService)
-        response = await service.get_remote_visitor_data(visitor_code, timeout)
-        return self.__handle_remote_visitor_data_response(response, visitor_code, add_data)
+        return await self._remote_data_manager.get_visitor_data(
+            visitor_code, add_data, data_filter, is_unique_identifier, timeout=timeout
+        )
 
-    def get_remote_visitor_data(self, visitor_code: str, add_data=True, timeout: Optional[float] = None) -> List[Data]:
+    def get_remote_visitor_data(
+        self,
+        visitor_code: str,
+        add_data=True,
+        timeout: Optional[float] = None,
+        data_filter: Optional[RemoteVisitorDataFilter] = None,
+        is_unique_identifier: bool = False,
+    ) -> List[Data]:
         """
         The get_remote_visitor_data is a synchronous method for retrieving custom data for
         the latest visit of `visitor_code` from Kameleoon Data API and optionally adding it
@@ -524,44 +558,24 @@ class KameleoonClient:
         :param timeout: requests Timeout for request (in seconds). Equals default_timeout in a config file.
         This field is optional.
         :type timeout: Optional[float]
+        :param data_filter: Filter that specifies which data should be retrieved from visits.
+        :type data_filter: RemoteVisitorDataFilter
+        :param is_unique_identifier: Parameter that specifies whether the visitorCode is a unique identifier.
+        :type is_unique_identifier: bool
 
         :return: A list of data assigned to the given visitor.
         :rtype: List[Data]
         """
-        service: DataService = self._network_manager.get_service(DataService)
-        coro = service.get_remote_visitor_data(visitor_code, timeout, sync=self.__is_sync_mode())
-        response = self.__make_sync_call_anyway(coro, "get_remote_visitor_data")
-        if response is None:
-            return []
-        return self.__handle_remote_visitor_data_response(response, visitor_code, add_data)
-
-    def __handle_remote_visitor_data_response(
-        self, response: Response, visitor_code: str, add_data: bool
-    ) -> List[Data]:
-        if response.content is None:
-            return []
-        data_list = self.__parse_custom_data_list(response.content)
-        if add_data:
-            self.add_data(visitor_code, *data_list)
-        return data_list
-
-    def __parse_custom_data_list(self, raw: Any) -> List[Data]:
-        try:
-            latest_record = raw.get("currentVisit")
-            if latest_record is None:
-                previous_visits = raw.get("previousVisits")
-                if previous_visits:
-                    latest_record = previous_visits[0]
-                else:
-                    return []
-            return [
-                CustomData(jcd.get("index", -1), *jcd.get("valuesCountMap", []))
-                for event in latest_record.get("customDataEvents", [])
-                if (jcd := event.get("data"))
-            ]
-        except Exception as ex:  # pylint: disable=W0703
-            self._config.logger.error(f"Parsing of visitor data failed: {ex}")
-            return []
+        coro = self._remote_data_manager.get_visitor_data(
+            visitor_code,
+            add_data,
+            data_filter,
+            is_unique_identifier,
+            self.__is_sync_mode(),
+            timeout,
+        )
+        result = self.__make_sync_call_anyway(coro, "get_remote_visitor_data")
+        return cast(List[Data], result)
 
     def get_visitor_warehouse_audience_async(
         self,
@@ -640,7 +654,11 @@ class KameleoonClient:
             VisitorCodeInvalid: Raise when the provided visitor code is not valid (empty, or longer than 255 characters)
         """
         coro = self._warehouse_manager.get_visitor_warehouse_audience(
-            visitor_code, custom_data_index, warehouse_key, timeout, sync=self.__is_sync_mode()
+            visitor_code,
+            custom_data_index,
+            warehouse_key,
+            timeout,
+            sync=self.__is_sync_mode(),
         )
         return self.__make_sync_call_anyway(coro, "get_visitor_warehouse_audience")
 
@@ -656,6 +674,7 @@ class KameleoonClient:
 
     def get_active_feature_list_for_visitor(self, visitor_code: str) -> List[str]:
         """
+        Depreacted function. Please use `get_active_features_for_visitor` instead.
         The get_active_feature_list_for_visitor method uses for obtaining a list of feature flag IDs:
         - currently targeted and active simultaneously for a visitor
 
@@ -667,10 +686,19 @@ class KameleoonClient:
         :rtype: List[int]
         """
 
+        warnings.warn(
+            "Call to deprecated function `get_active_feature_list_for_visitor`."
+            "Please use `get_active_features` instead.",
+            category=DeprecationWarning,
+            stacklevel=2,
+        )
+
         def filter_conditions(feature_flag: FeatureFlag) -> bool:
+            if not feature_flag.environment_enabled:
+                return False
             (variation, rule) = self.__calculate_variation_rule_for_feature(visitor_code, feature_flag)
             variation_key = self.__calculate_variation_key(variation, rule, feature_flag)
-            return variation_key != Variation.Type.OFF.value
+            return variation_key != configuration.Variation.Type.OFF.value
 
         return list(
             map(
@@ -682,7 +710,45 @@ class KameleoonClient:
             )
         )
 
-    def get_feature_variation_key(self, visitor_code: str, feature_key: str) -> str:
+    def get_active_features(self, visitor_code: str) -> Dict[str, Variation]:
+        """
+        The get_active_features method uses for obtaining a information about the active feature flags
+         that are available for the visitor.
+
+        :param visitor_code: unique identifier of a visitor
+        :type visitor_code: str
+
+        :return: Dictionary that contains the assigned variations of the active features using the keys
+         of the corresponding active features.
+        :rtype: Dict[str, Variation]
+
+        :raises:
+            VisitorCodeInvalid: Raise when the provided visitor code is not valid
+        """
+
+        validate_visitor_code(visitor_code)
+        map_active_features: Dict[str, Variation] = {}
+        for feature_flag in self._data_file.feature_flags.values():
+            if not feature_flag.environment_enabled:
+                continue
+            (var_by_exp, rule) = self.__calculate_variation_rule_for_feature(visitor_code, feature_flag)
+            variation_key = self.__calculate_variation_key(var_by_exp, rule, feature_flag)
+            if variation_key == configuration.Variation.Type.OFF.value:
+                continue
+            variation = feature_flag.get_variation(variation_key)
+            variables: Dict[str, Variable] = {}
+            if variation is not None:
+                for variable in variation.variables:
+                    variables[variable.key] = Variable(variable.key, variable.get_type(), variable.get_value())
+            map_active_features[feature_flag.feature_key] = Variation(
+                variation_key,
+                var_by_exp.variation_id if var_by_exp is not None else None,
+                rule.experiment_id if rule is not None else None,
+                variables,
+            )
+        return map_active_features
+
+    def get_feature_variation_key(self, visitor_code: str, feature_key: str, is_unique_identifier: bool = False) -> str:
         """
         Returns a variation key for visitor code
 
@@ -696,6 +762,8 @@ class KameleoonClient:
         :type visitor_code: str
         :param feature_key: unique identifier of feature flag
         :type feature_key: str
+        :param is_unique_identifier: Parameter that specifies whether the visitorCode is a unique identifier.
+        :type is_unique_identifier: bool
 
         :return: Returns a variation key for visitor code
         :rtype: str
@@ -711,11 +779,11 @@ class KameleoonClient:
             FeatureEnvironmentDisabled: Exception indicating that feature flag is disabled for the
                                         visitor's current environment.
         """
-        (_, variation_key) = self.__get_feature_variation_key(visitor_code, feature_key)
+        (_, variation_key) = self.__get_feature_variation_key(visitor_code, feature_key, is_unique_identifier)
         return variation_key
 
     def get_feature_variable(
-        self, visitor_code: str, feature_key: str, variable_key: str
+        self, visitor_code: str, feature_key: str, variable_key: str, is_unique_identifier: bool = False
     ) -> Union[bool, str, float, Dict[str, Any], List[Any], None]:
         """
         Retrieves a feature variable value from assigned for visitor variation
@@ -727,6 +795,8 @@ class KameleoonClient:
         :type feature_key: str
         :param variable_name: variable name you want to retrieve
         :type variable_name: str
+        :param is_unique_identifier: Parameter that specifies whether the visitorCode is a unique identifier.
+        :type is_unique_identifier: bool
 
         :return: Feature variable value from assigned for visitor variation
         :rtype: Union[bool, str, float, Dict, List]
@@ -744,7 +814,9 @@ class KameleoonClient:
                                         visitor's current environment.
         """
 
-        (feature_flag, variation_key) = self.__get_feature_variation_key(visitor_code, feature_key)
+        (feature_flag, variation_key) = self.__get_feature_variation_key(
+            visitor_code, feature_key, is_unique_identifier
+        )
         variation = feature_flag.get_variation(variation_key)
         variable = variation.get_variable_by_key(variable_key) if variation else None
         if variable is None:
@@ -773,55 +845,8 @@ class KameleoonClient:
     #     return None
 
     # pylint: disable=W0238
-    def __check_targeting(self, visitor_code: str, campaign: Dict[str, Any]):
-        if (SEGMENT not in campaign) or (campaign[SEGMENT] is None):
-            return True
-        visitor = self._visitor_manager.get_visitor(visitor_code)
-        campaign_id = campaign["id"]
-        return campaign[SEGMENT].check_tree(
-            lambda type: self.__get_condition_data(type, visitor, visitor_code, campaign_id)
-        )
-
-    def __check_targeting_id_object(self, visitor_code: str, campaign_id: int, targeting_object: Rule) -> bool:
-        if targeting_object.targeting_segment is None:
-            return True
-        visitor = self._visitor_manager.get_visitor(visitor_code)
-        return (
-            targeting_object.targeting_segment.check_tree(
-                lambda type: self.__get_condition_data(type, visitor, visitor_code, campaign_id)
-            )
-            is True
-        )
-
-    # pylint: disable=R0911
-    @staticmethod
-    def __get_condition_data(
-        condition_type_literal: str, visitor: Optional[Visitor], visitor_code: str, campaign_id: int
-    ):
-        # pylint: disable=E1101,W0212
-        condition_type = enum_from_literal(
-            condition_type_literal, TargetingConditionType, TargetingConditionType.UNKNOWN
-        )
-        if visitor:
-            if condition_type == TargetingConditionType.BROWSER:
-                return visitor.browser
-            if condition_type == TargetingConditionType.DEVICE_TYPE:
-                return visitor.device
-            if condition_type == TargetingConditionType.CUSTOM_DATUM:
-                return visitor.custom_data
-            if condition_type == TargetingConditionType.CONVERSIONS:
-                return visitor.conversions
-            if condition_type in (TargetingConditionType.PAGE_URL, TargetingConditionType.PAGE_TITLE):
-                return visitor.page_view_visits
-            if condition_type == TargetingConditionType.TARGET_EXPERIMENT:
-                return visitor.variations
-            if condition_type == TargetingConditionType.EXCLUSIVE_EXPERIMENT:
-                return campaign_id, visitor.variations
-        if condition_type == TargetingConditionType.VISITOR_CODE:
-            return visitor_code
-        if condition_type == TargetingConditionType.SDK_LANGUAGE:
-            return SdkLanguageCondition.SdkInfo(SdkVersion.NAME, SdkVersion.VERSION)
-        return None
+    def __check_targeting(self, visitor_code: str, campaign_id: int, rule: Rule):
+        return self._targeting_manager.check_targeting(visitor_code, campaign_id, rule)
 
     # pylint: disable=W0105
     """
@@ -861,11 +886,13 @@ class KameleoonClient:
         try:
             configuration_json = self._obtain_configuration(time_stamp)
             if configuration_json:
-                self._data_file = DataFile.from_json(self._config.environment, configuration_json)
+                self._data_file = DataFile.from_json(self._config.environment, configuration_json, self._config.logger)
                 self._cookie_manager.consent_required = (
                     self._data_file.settings.is_consent_required and not self._data_file.has_any_targeted_delivery_rule
                 )
                 self._network_manager.url_provider.apply_data_api_domain(self._data_file.settings.data_api_domain)
+                self._targeting_manager = TargetingManager(self._visitor_manager, self._data_file)
+                self._visitor_manager.custom_data_info = self._data_file.custom_data_info
                 self._call_update_handler_if_needed(time_stamp is not None)
             success = True
         except Exception as ex:
@@ -1011,9 +1038,12 @@ class KameleoonClient:
             return False
         return not has_running_event_loop()
 
-    def __send_tracking_request(self, visitor_code: str, visitor: Optional[Visitor], force_request=True) -> None:
+    def __send_tracking_request(
+        self, visitor_code: str, visitor: Optional[Visitor], force_request=True, is_unique_identifier=False
+    ) -> None:
         if (visitor is None) and self._data_file.settings.is_consent_required:
             return
+        use_mapping_value, visitor = self.__create_anonymous_if_required(visitor_code, visitor, is_unique_identifier)
         if visitor:
             consent = not self._data_file.settings.is_consent_required or visitor.legal_consent
             user_agent = visitor.user_agent
@@ -1027,7 +1057,9 @@ class KameleoonClient:
                 return
             unsent.append(ActivityEvent())
         service: DataService = self._network_manager.get_service(DataService)
-        coro = service.send_tracking_data(visitor_code, unsent, user_agent, sync=self.__is_sync_mode())
+        coro = service.send_tracking_data(
+            visitor_code, unsent, user_agent, is_unique_identifier=use_mapping_value, sync=self.__is_sync_mode()
+        )
 
         async def call() -> Response:
             response = await coro
@@ -1038,6 +1070,21 @@ class KameleoonClient:
 
         self.__run_call(call())
 
+    def __create_anonymous_if_required(
+        self, visitor_code: str, visitor: Optional[Visitor], is_unique_identifier: bool
+    ) -> tuple[bool, Optional[Visitor]]:
+        use_mapping_value = is_unique_identifier and (visitor is not None and visitor.mapping_identifier is not None)
+        # need to find if anonymous visitor is behind unique (anonym doesn't exist if MappingIdentifier == null)
+        if is_unique_identifier and (visitor is None or visitor.mapping_identifier is None):
+            # We haven't anonymous behind, in this case we should create "fake" anonymous with id == visitorCode
+            # and link it with with mapping value == visitorCode (like we do as we have real anonymous visitor)
+            mapping_index = self._data_file.custom_data_info.mapping_identifier_index
+            if mapping_index is not None:
+                visitor = self._visitor_manager.add_data(
+                    visitor_code, CustomData(mapping_index, visitor_code), logger=self._config.logger
+                )
+        return use_mapping_value, visitor
+
     @staticmethod
     def __select_unsent_data(visitor: Visitor, consent: bool) -> List[Sendable]:
         if consent:
@@ -1045,28 +1092,29 @@ class KameleoonClient:
         tdr_variations = (av for av in visitor.variations.values() if av.rule_type == RuleType.TARGETED_DELIVERY)
         return [sd for sd in itertools.chain(visitor.conversions, tdr_variations) if not sd.sent]
 
-    def __get_feature_variation_key(self, visitor_code: str, feature_key: str) -> Tuple[FeatureFlag, str]:
+    def __get_feature_variation_key(
+        self, visitor_code: str, feature_key: str, is_unique_identifier: bool = False
+    ) -> Tuple[FeatureFlag, str]:
         """
         helper method for getting variation key for feature flag
         """
-        validate_visitor_code(visitor_code)
         feature_flag = self._data_file.get_feature_flag(feature_key)
         (variation, rule) = self.__calculate_variation_rule_for_feature(visitor_code, feature_flag)
         variation_key = self.__calculate_variation_key(variation, rule, feature_flag)
-        visitor = self.__assign_feature_variation(visitor_code, rule, variation)
-        self.__send_tracking_request(visitor_code, visitor)
+        self.__assign_feature_variation(visitor_code, rule, variation)
+        self.flush(visitor_code, is_unique_identifier)
         return (feature_flag, variation_key)
 
     def __assign_feature_variation(
         self, visitor_code: str, rule: Optional[Rule], variation: Optional[VariationByExposition]
-    ) -> Optional[Visitor]:
-        visitor = self._visitor_manager.get_visitor(visitor_code)
-        if rule and variation and (variation.variation_id is not None):
-            if visitor is None:
-                visitor = self._visitor_manager.get_or_create_visitor(visitor_code)
-            as_variation = AssignedVariation(rule.experiment_id, variation.variation_id, rule.type)
-            visitor.assign_variation(as_variation)
-        return visitor
+    ) -> None:
+        if rule is not None:
+            visitor = self._visitor_manager.get_or_create_visitor(visitor_code)
+            experiment_id = rule.experiment_id
+            variation_id = variation.variation_id if variation is not None else None
+            if experiment_id is not None and variation_id is not None:
+                as_variation = AssignedVariation(experiment_id, variation_id, rule.type)
+                visitor.assign_variation(as_variation)
 
     def __calculate_variation_rule_for_feature(
         self, visitor_code: str, feature_flag: FeatureFlag
@@ -1074,24 +1122,31 @@ class KameleoonClient:
         """helper method for calculate variation key for feature flag"""
         for rule in feature_flag.rules:
             # check if visitor is targeted for rule, else next rule
-            if not self.__check_targeting_id_object(visitor_code, feature_flag.id_, rule):
+            if not self.__check_targeting(visitor_code, rule.experiment_id, rule):
                 continue
+            visitor = self._visitor_manager.get_visitor(visitor_code)
+            # use mappingIdentifier instead of visitorCode if it was set up
+            code_for_hash = (
+                visitor.mapping_identifier
+                if visitor is not None and visitor.mapping_identifier is not None
+                else visitor_code
+            )
             # uses for rule exposition
-            hash_rule = obtain_hash_double_rule(visitor_code, rule.id_, rule.respool_time)
+            hash_rule = obtain_hash_double_rule(code_for_hash, rule.id_, rule.respool_time)
             # check main expostion for rule with hashRule
             if hash_rule <= rule.exposition:
-                if rule.is_targeted_delivery and len(rule.variation_by_exposition) > 0:
-                    return (rule.variation_by_exposition[0], rule)
+                if rule.is_targeted_delivery:
+                    return (rule.first_variation, rule)
 
                 # uses for variation's expositions
-                hash_variation = obtain_hash_double_rule(visitor_code, rule.experiment_id, rule.respool_time)
+                hash_variation = obtain_hash_double_rule(code_for_hash, rule.experiment_id, rule.respool_time)
                 # get variation with hash_variation
                 variation = rule.get_variation(hash_variation)
                 if variation:
                     return (variation, rule)
             elif rule.is_targeted_delivery:
                 break
-        return (None, None)
+        return None, None
 
     @staticmethod
     def __calculate_variation_key(
@@ -1102,7 +1157,7 @@ class KameleoonClient:
         if var_by_exp:
             return var_by_exp.variation_key
         if rule and rule.is_experimentation:
-            return Variation.Type.OFF.value
+            return configuration.Variation.Type.OFF.value
         return feature_flag.default_variation_key
 
     def _is_consent_given(self, visitor_code: str) -> bool:
